////////////////////////////////////
//  RainyDay.java
//
//  This code was created for the 3nd Java Applet Coding Contest at http://www.cfxweb.net/
//
//  If anybody can find something usefull in this code please use it and let me know.
//
//  Bob.
//  hk@dgmr.nl
//  http://home.kabelfoon.nl/~rhordijk
///////////////////////////////////////
package rainyday;

public class NoiseMap
{
  private int[] P = new int[0x202];
  private int[][] G2I = new int[0x202][2];
  private int[][] G3I = new int[0x202][3];
  private int B = 0x100;

  private int width = 0;
  private int height = 0;
  private int noiseoffset = 0;

  public int Map[];

  private void initnoise()
  {
    double G20, G21, l;

    for(int i = 0; i < B ; i++){
      P[i] = i;
      G20 = Math.floor((Math.random() * 2 * B) - B)/B;
      G21 = Math.floor((Math.random() * 2 * B) - B)/B;
      l = Math.sqrt(G20 * G20 + G21 * G21);
      if (l > 1e-5) {
        G20 /= l;
        G21 /= l;
      }
      G2I[i][0] = (int) Math.round(G20 * 1024);
      G2I[i][1] = (int) Math.round(G21 * 1024);

      double G30 = Math.floor((Math.random() * 2 * B) - B)/B;
      double G31 = Math.floor((Math.random() * 2 * B) - B)/B;
      double G32 = Math.floor((Math.random() * 2 * B) - B)/B;

      l = Math.sqrt(G30 * G30 + G31 * G31 + G32 * G32);
      if (l > 1e-5) {
        G30 /= l;
        G31 /= l;
        G32 /= l;
      }
      G3I[i][0] = (int) Math.round(G30 * 1024);
      G3I[i][1] = (int) Math.round(G31 * 1024);
      G3I[i][2] = (int) Math.round(G32 * 1024);
    }

    int swap, temp;
    for(int i = 0; i < B ; i++){
      swap = (int) Math.round(Math.random() * B);
      temp    = P[i];
      P[i]    = P[swap];
      P[swap] = temp;
    }

    for(int i = 0; i < B + 2 ; i++){
      P[B + i]   = P[i];
      G2I[B + i][0] = G2I[i][0];
      G2I[B + i][1] = G2I[i][1];

      G3I[B + i][0] = G3I[i][0];
      G3I[B + i][1] = G3I[i][1];
      G3I[B + i][2] = G3I[i][2];
    }
  }

  private int noise2Di(int x, int y)
  {
    int t = x + 0x40000000;
    int bx0 = (t >> 10) & 0xFF;
    int bx1 = (bx0 + 1) & 0xFF;
    int i  =  P[bx0];
    int j  =  P[bx1];
    int rx0 = t & 1023;
    int rx1 = rx0 - 1024;

    t = y + 0x40000000;
    int by0 = (t >> 10) & 0xFF;
    int by1 = (by0 + 1) & 0xFF;
    int ry0 = t & 1023;
    int ry1 = ry0 - 1024;

    int b00 = P[i + by0];
    int b10 = P[j + by0];
    int b01 = P[i + by1];
    int b11 = P[j + by1];

    int sx = (rx0 * rx0 * (3072 - 2 * rx0)) >> 20;
    int sy = (ry0 * ry0 * (3072 - 2 * ry0)) >> 20;

    int u = (rx0 * G2I[b00][0] + ry0 * G2I[b00][1]);
    int v = (rx1 * G2I[b10][0] + ry0 * G2I[b10][1]);
    int a = (u << 10) + sx * (v - u);
    a = a >> 20;

    u = (rx0 * G2I[b01][0] + ry1 * G2I[b01][1]);
    v = (rx1 * G2I[b11][0] + ry1 * G2I[b11][1]);
    int b = (u << 10) + sx * (v - u);
    b = b >> 20;

    int r = (a  * 1024)  + sy * (b - a);
    return  r >> 10;
  }

  private int noise3Di(int x, int y, int z)
  {
    int t = x + 0x40000000;
    int bx0 = (t >> 10) & 0xFF;
    int bx1 = (bx0 + 1) & 0xFF;
    int i  =  P[bx0];
    int j  =  P[bx1];
    int rx0 = t & 1023;
    int rx1 = rx0 - 1024;

    t = y + 0x40000000;
    int by0 = (t >> 10) & 0xFF;
    int by1 = (by0 + 1) & 0xFF;
    int ry0 = t & 1023;
    int ry1 = ry0 - 1024;

    t = z + 0x40000000;
    int bz0 = (t >> 10) & 0xFF;
    int bz1 = (bz0 + 1) & 0xFF;
    int rz0 = t & 1023;
    int rz1 = rz0 - 1024;

    int b00 = P[i + by0];
    int b10 = P[j + by0];
    int b01 = P[i + by1];
    int b11 = P[j + by1];

    int sx = (rx0 * rx0 * (3072 - 2 * rx0)) >> 20;
    int sy = (ry0 * ry0 * (3072 - 2 * ry0)) >> 20;
    int sz = (rz0 * rz0 * (3072 - 2 * rz0)) >> 20;

    int u = (rx0 * G3I[b00 + bz0][0] + ry0 * G3I[b00 + bz0][1] + rz0 * G3I[b00 + bz0][2]);
    int v = (rx1 * G3I[b10 + bz0][0] + ry0 * G3I[b10 + bz0][1] + rz0 * G3I[b10 + bz0][2]);
    int a = (u << 10) + sx * (v - u);
    a = a >> 20;

    u = (rx0 * G3I[b01 + bz0][0] + ry1 * G3I[b01 + bz0][1] + rz0 * G3I[b01 + bz0][2]);
    v = (rx1 * G3I[b11 + bz0][0] + ry1 * G3I[b11 + bz0][1] + rz0 * G3I[b11 + bz0][2]);
    int b = (u << 10) + sx * (v - u);
    b = b >> 20;

    int c = (a << 10) + sy * (b - a);

    u = (rx0 * G3I[b00 + bz1][0] + ry0 * G3I[b00 + bz1][1] + rz1 * G3I[b00 + bz1][2]);
    v = (rx1 * G3I[b10 + bz1][0] + ry0 * G3I[b10 + bz1][1] + rz1 * G3I[b10 + bz1][2]);
    a = (u << 10) + sx * (v - u);
    a = a  >> 20;

    u = (rx0 * G3I[b01 + bz1][0] + ry1 * G3I[b01 + bz1][1] + rz1 * G3I[b01 + bz1][2]);
    v = (rx1 * G3I[b11 + bz1][0] + ry1 * G3I[b11 + bz1][1] + rz1 * G3I[b11 + bz1][2]);
    b = (u << 10) + sx * (v - u);
    b = b >> 20;

    int d = (a << 10) + sy * (b - a);

    int r = (c << 10)  + sz * (d - c);
    return  r >> 20;
  }

  public int turbulence2Di(int x, int y,  int n)
  {
    int r = 0;
    int a = 1;
    for(int i = n - 1; i >= 0 ; i --) {
      r += noise2Di(x,y) / a;
      x = x * 2;
      y = y * 2;
      a = a * 2;
    }
    return r;
  }

  public int turbulence3Di(int x, int y, int z, int n)
  {
    int r = 0;
    int a = 1;
    for(int i = n - 1; i >= 0 ; i --) {
      r += noise3Di(x,y,z) / a;
      x = x << 1;
      y = y << 1;
      z = z << 1;
      a = a << 1;
    }
    return r;
  }


  public void UpdateNoiseMap(int lines)
  {
    System.arraycopy(Map, 0, Map, lines * width, (height - lines ) * width);

    noiseoffset -= 100 * lines;
    for ( int y = 0; y < lines; y++) {
      for ( int x = 0; x < width; x++) {
         Map[y * width + x] = (int) Math.round(255 * Math.pow(Math.abs(turbulence2Di(x * 100, noiseoffset + y * 100, 4) / 4)/255.0, 1.87));
      }
    }
  }


  public NoiseMap(int w, int h)
  {
    initnoise();

    width = 2 * w;
    height = h;

    Map = new int[width * (2 * height)];
  }
}